import pytest
import torch
import torch.nn as nn

from ideal_words import FactorEmbedding, IdealWords


def test_auxillary_setup():
    # toy example to test the auxillary attributes used during ideal word computation
    Z1 = ['X1', 'X2']
    Z2 = ['Y1', 'Y2', 'Y3', 'Y4']
    Z3 = ['Z1', 'Z2', 'Z3']

    # dummy embeddings
    torch.manual_seed(1337)
    embeddings = torch.randn(len(Z1) * len(Z2) * len(Z3), 4)

    # we use random embeddings here as we only want to verify auxillary attributes used during ideal word computation
    txt_encoder = nn.Identity()

    def tokenizer(text: list[str]) -> torch.Tensor:
        _ = text
        return embeddings

    # when device is not specified, it defaults to cuda if available
    fe = FactorEmbedding(txt_encoder, tokenizer, device=None)
    assert fe.device == 'cuda' if torch.cuda.is_available() else 'cpu'

    # create factor embeddings and compute ideal words
    fe = FactorEmbedding(txt_encoder, tokenizer, device='cpu')
    iw = IdealWords(fe, [Z1, Z2, Z3])

    # factor combinations
    assert iw.pairs == [
        ('X1', 'Y1', 'Z1'),
        ('X1', 'Y1', 'Z2'),
        ('X1', 'Y1', 'Z3'),
        ('X1', 'Y2', 'Z1'),
        ('X1', 'Y2', 'Z2'),
        ('X1', 'Y2', 'Z3'),
        ('X1', 'Y3', 'Z1'),
        ('X1', 'Y3', 'Z2'),
        ('X1', 'Y3', 'Z3'),
        ('X1', 'Y4', 'Z1'),
        ('X1', 'Y4', 'Z2'),
        ('X1', 'Y4', 'Z3'),
        ('X2', 'Y1', 'Z1'),
        ('X2', 'Y1', 'Z2'),
        ('X2', 'Y1', 'Z3'),
        ('X2', 'Y2', 'Z1'),
        ('X2', 'Y2', 'Z2'),
        ('X2', 'Y2', 'Z3'),
        ('X2', 'Y3', 'Z1'),
        ('X2', 'Y3', 'Z2'),
        ('X2', 'Y3', 'Z3'),
        ('X2', 'Y4', 'Z1'),
        ('X2', 'Y4', 'Z2'),
        ('X2', 'Y4', 'Z3'),
    ]

    # factor to index mapping
    assert iw.factor2idx == {
        'X1': (0, 0),
        'X2': (0, 1),
        'Y1': (1, 0),
        'Y2': (1, 1),
        'Y3': (1, 2),
        'Y4': (1, 3),
        'Z1': (2, 0),
        'Z2': (2, 1),
        'Z3': (2, 2),
    }

    for zi in iw.factor2idx:
        i, j = iw.factor2idx[zi]
        assert iw.factors[i][j] == zi

    # automatic uniform weights
    assert len(iw.weights) == 3
    assert iw.weights[0] == [1 / 2] * 2
    assert iw.weights[1] == [1 / 4] * 4
    assert iw.weights[2] == [1 / 3] * 3

    assert iw.factor_indices == [
        {
            'X1': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
            'X2': [12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23],
        },
        {
            'Y1': [0, 1, 2, 12, 13, 14],
            'Y2': [3, 4, 5, 15, 16, 17],
            'Y3': [6, 7, 8, 18, 19, 20],
            'Y4': [9, 10, 11, 21, 22, 23],
        },
        {
            'Z1': [0, 3, 6, 9, 12, 15, 18, 21],
            'Z2': [1, 4, 7, 10, 13, 16, 19, 22],
            'Z3': [2, 5, 8, 11, 14, 17, 20, 23],
        },
    ]

    for i, factor_index in enumerate(iw.factor_indices):
        for zi, inds in factor_index.items():
            # verify that the returned indices actually map back to the expected zi
            for ind in inds:
                assert iw.pairs[ind][i] == zi

            # compare mean over indexed embedding vectors to ideal word
            assert torch.allclose(iw.get_iw(zi), iw.embeddings[inds].mean(dim=0) - iw.u_zero)

    # u_zero is just the mean over all embeddings
    assert torch.allclose(iw.u_zero, iw.embeddings.mean(dim=0))


def test_compositional_toy_embeddings():
    # toy example with predefined embeddings that are already compositional
    Z1 = ['blue', 'red']
    Z2 = ['car', 'bike']

    # joint representations for ideal words
    joint_embeddings = {
        'blue car': torch.Tensor([1, 0, 1, 0]),
        'red car': torch.Tensor([0, 1, 1, 0]),
        'blue bike': torch.Tensor([1, 0, 0, 1]),
        'red bike': torch.Tensor([0, 1, 0, 1]),
    }

    # single representations for real words
    single_embeddings = {
        'blue': torch.Tensor([1, 0, 0, 0]),
        'red': torch.Tensor([0, 1, 0, 0]),
        'car': torch.Tensor([0, 0, 1, 0]),
        'bike': torch.Tensor([0, 0, 0, 1]),
    }

    # embeddings have 2 dimensions per factor and use one-hot encoding
    embeddings = {**joint_embeddings, **single_embeddings}

    # we use the predefined embeddings by returning them from the tokenizer and using nn.Identity() as text encoder
    txt_encoder = nn.Identity()

    def tokenizer(text: list[str]) -> torch.Tensor:
        return torch.stack([embeddings[z] for z in text])

    # create factor embeddings and compute ideal words
    fe = FactorEmbedding(txt_encoder, tokenizer, normalize=False, device='cpu')
    iw = IdealWords(fe, [Z1, Z2])

    # test ideal word computation
    assert torch.allclose(iw.u_zero, torch.Tensor([0.5, 0.5, 0.5, 0.5]))
    assert torch.allclose(iw.get_iw('blue'), torch.Tensor([0.5, -0.5, 0, 0]))
    assert torch.allclose(iw.get_iw('red'), torch.Tensor([-0.5, 0.5, 0, 0]))
    assert torch.allclose(iw.get_iw('car'), torch.Tensor([0, 0, 0.5, -0.5]))
    assert torch.allclose(iw.get_iw('bike'), torch.Tensor([0, 0, -0.5, 0.5]))

    # approximations are perfect because embeddings were already compositional, thus distance is 0
    assert iw.iw_score == (0.0, 0.0)
    assert iw.iw_accuracy == 1.0

    # ideal word approximations exactly equal the original embeddings
    for caption, embedding in joint_embeddings.items():
        z = caption.split(' ')
        assert torch.allclose(iw.get_uz(z), embedding)  # u_z = u_zero + u_color + u_object

    # real words are just the embeddings themselves
    for caption, embedding in single_embeddings.items():
        assert torch.allclose(iw.get_rw(caption), embedding)

    # real word approximations are also close to the original embeddings except scaled by 0.5
    assert torch.allclose(iw.get_uz(['blue', 'car'], approx='real'), torch.Tensor([0.5, 0, 0.5, 0]))
    assert torch.allclose(iw.get_uz(['red', 'car'], approx='real'), torch.Tensor([0, 0.5, 0.5, 0]))
    assert torch.allclose(iw.get_uz(['blue', 'bike'], approx='real'), torch.Tensor([0.5, 0, 0, 0.5]))
    assert torch.allclose(iw.get_uz(['red', 'bike'], approx='real'), torch.Tensor([0, 0.5, 0, 0.5]))

    # real word approximations are not perfect even when embeddings are compositional, thus distance greater than 0
    assert iw.rw_score == (pytest.approx(0.5), 0.0)
    assert iw.rw_accuracy == 1.0

    # requesting approximation modes other than 'ideal' or 'real' raises a value error
    with pytest.raises(ValueError):
        iw.get_uz(['blue', 'car'], approx='foobar')


def test_noncompositional_toy_embeddings():
    # toy example with predefined embeddings that are not compositional
    Z1 = ['blue', 'red']
    Z2 = ['car', 'bike']

    # embeddings use one-hot encoding
    embeddings = {
        'blue car': torch.Tensor([1, 0, 0, 0]),
        'red car': torch.Tensor([0, 1, 0, 0]),
        'blue bike': torch.Tensor([0, 0, 1, 0]),
        'red bike': torch.Tensor([0, 0, 0, 1]),
    }

    # we use the predefined embeddings by returning them from the tokenizer and using nn.Identity() as text encoder
    txt_encoder = nn.Identity()

    def tokenizer(text: list[str]) -> torch.Tensor:
        return torch.stack([embeddings[z] for z in text])

    # create factor embeddings and compute ideal words
    fe = FactorEmbedding(txt_encoder, tokenizer, normalize=False, device='cpu')
    iw = IdealWords(fe, [Z1, Z2])

    assert torch.allclose(iw.u_zero, torch.Tensor([0.25, 0.25, 0.25, 0.25]))
    assert torch.allclose(iw.get_iw('blue'), torch.Tensor([0.25, -0.25, 0.25, -0.25]))
    assert torch.allclose(iw.get_iw('red'), torch.Tensor([-0.25, 0.25, -0.25, 0.25]))
    assert torch.allclose(iw.get_iw('car'), torch.Tensor([0.25, 0.25, -0.25, -0.25]))
    assert torch.allclose(iw.get_iw('bike'), torch.Tensor([-0.25, -0.25, 0.25, 0.25]))

    # in this case the ideal word approximations do not match the original embeddings
    assert torch.allclose(iw.get_uz(('blue', 'car')), torch.Tensor([0.75, 0.25, 0.25, -0.25]))
    assert torch.allclose(iw.get_uz(('red', 'car')), torch.Tensor([0.25, 0.75, -0.25, 0.25]))
    assert torch.allclose(iw.get_uz(('blue', 'bike')), torch.Tensor([0.25, -0.25, 0.75, 0.25]))
    assert torch.allclose(iw.get_uz(('red', 'bike')), torch.Tensor([-0.25, 0.25, 0.25, 0.75]))

    # approximations are imperfect but distances are always the same so we have score > 0 but std = 0
    assert iw.iw_score == (0.25, 0.0)
    assert iw.iw_accuracy == 1.0  # approximations are still close enough


def test_random_embeddings():
    # toy example with randomly initialized embeddings
    Z1 = ['A', 'B', 'C', 'D', 'E']
    Z2 = ['1', '2', '3', '4', '5']
    Z3 = ['.', '?', '!', ',', ';']

    # we have 5 x 5 x 5 = 125 different combinations and an embedding dimension of 64
    torch.manual_seed(1337)
    embeddings = torch.randn(125, 64)

    # we use the predefined embeddings by returning them from the tokenizer and using nn.Identity() as text encoder
    txt_encoder = nn.Identity()

    def tokenizer(text: list[str]) -> torch.Tensor:
        _ = text
        return embeddings

    # create factor embeddings and compute ideal words
    fe = FactorEmbedding(txt_encoder, tokenizer, normalize=True, device='cpu')
    iw = IdealWords(fe, [Z1, Z2, Z3])

    # mean over normally distributed random vectors is close to zero
    assert iw.u_zero.norm() <= 0.1

    for iw_per_factor in iw.ideal_words:
        # ideal words belonging to a factor Z_i should sum to zero
        assert torch.allclose(iw_per_factor.sum(dim=0), torch.zeros(64))

    # approximations are not perfect because embeddings are random
    assert iw.iw_score[0] >= 0.0
    assert iw.iw_score[1] >= 0.0

    assert iw.avg_score[0] >= 0.0
    assert iw.avg_score[1] >= 0.0



import argparse
import os
import time

import torch
from open_clip import create_model_and_transforms, get_tokenizer

from ideal_words import FactorEmbedding, IdealWords


class AttributeObjectFactorEmbedding(FactorEmbedding):
    def encode_text(self, text: torch.Tensor) -> torch.Tensor:
        # CLIP is not only a text encoder, so we need to specify how to use it for encoding text
        return self.txt_encoder.encode_text(text)

    def joint_repr(self, pair: tuple[str, ...]) -> str:
        attr, obj = pair
        # classic zero shot type caption of attribute-object dataset
        return f'an image of a {attr} {obj}'

    def single_repr(self, zi: str, factor_idx: int) -> str:
        # for the real word score, we also need to encode factors separately
        if factor_idx == 0:
            # zi is an attribute
            return f'image of a {zi} object'
        elif factor_idx == 1:
            # zi is an object type
            return f'image of a {zi}'
        else:
            raise IndexError(f'Invalid factor index: {factor_idx}')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--random-init', action='store_true', help='use a randomly initialized encoder instead of the pretrained one'
    )
    parser.add_argument(
        '--score-mode',
        type=str,
        default='paper_repro',
        choices=['avg_dist', 'avg_sq_dist', 'paper_repro'],
        help='customize how scores are computed, see `IdealWords` docstring for details',
    )
    args = parser.parse_args()

    # keep track of execution time
    start = time.time()

    # in the paper, they used a ViT-L-14 based CLIP model from OpenAI
    pretrained = None if args.random_init else 'openai'
    clip, *_ = create_model_and_transforms('ViT-L-14', precision='fp16', pretrained=pretrained)
    tokenizer = get_tokenizer('ViT-L-14')
    print('Loaded CLIP model and tokenizer.')

    # load factors from mit-states
    dirname = os.path.dirname(os.path.abspath(__file__))
    factors = {}
    with open(os.path.join(dirname, 'mit-states.csv'), 'r') as f:
        Z1, Z2 = [line.strip().split(',') for line in f.readlines()]
        factors['mit-states'] = Z1, Z2
    with open(os.path.join(dirname, 'ut-zappos.csv'), 'r') as f:
        Z1, Z2 = [line.strip().split(',') for line in f.readlines()]
        factors['ut-zappos'] = Z1, Z2
    print('Loaded factors for MIT-States and UT Zappos.')

    fe = AttributeObjectFactorEmbedding(clip, tokenizer, normalize=True)

    # compute ideal words and score for mit-states
    mit_iw = IdealWords(fe, factors['mit-states'], weights=None, score_mode=args.score_mode)
    mit_iw_score, mit_iw_std = mit_iw.iw_score
    mit_rw_score, mit_rw_std = mit_iw.rw_score
    mit_avg_score, mit_avg_std = mit_iw.avg_score
    print('Computed ideal words and scores for MIT-States.')

    # compute ideal words and score for ut-zappos
    ut_iw = IdealWords(fe, factors['ut-zappos'], weights=None, score_mode=args.score_mode)
    ut_iw_score, ut_iw_std = ut_iw.iw_score
    ut_rw_score, ut_rw_std = ut_iw.rw_score
    ut_avg_score, ut_avg_std = ut_iw.avg_score
    print('Computed ideal words and scores for UT Zappos.')

    # print table
    print()
    print('                 IW            RW            Avg    ')
    print('----------------------------------------------------')
    print(
        f'MIT-States   {mit_iw_score:.2f} ± {mit_iw_std:.2f}   {mit_rw_score:.2f} ± {mit_rw_std:.2f}   '
        f'{mit_avg_score:.2f} ± {mit_avg_std:.2f}'
    )
    print(
        f'UT Zappos    {ut_iw_score:.2f} ± {ut_iw_std:.2f}   {ut_rw_score:.2f} ± {ut_rw_std:.2f}   '
        f'{ut_avg_score:.2f} ± {ut_avg_std:.2f}'
    )
    print(f'\nTook {time.time() - start:.2f} seconds')



from functools import reduce
from itertools import product
from math import isclose
from typing import Callable, Literal, Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F
import tqdm
from torch.utils.data import DataLoader, TensorDataset


class FactorEmbedding:
    """
    Utility class that encapsulates the some additional logic needed to compute ideal words and scores. That is, how to
    compute text embeddings (e.g, a tokenizer and a text encoder), how to jointly represent a tuple (z_1, ..., z_k) as a
    string, and how to represent a single z_i as a string.

    By default, text is embedded using the given tokenizer and the forward method of the given text encoder, a tuple z
    is represented as a string by joining tuple elements with spaces, and a single zi is its own string representation.
    If you want to customize this behaviour, sub-type this class.
    """

    def __init__(
        self,
        txt_encoder: nn.Module,
        tokenizer: Callable[[list[str]], torch.Tensor],
        normalize: bool = True,
        batch_size: int = 64,
        device: str | None = None,
    ) -> None:
        """
        Parameters
        ----------
            txt_encoder: Text encoder for computing the text embeddings.

            tokenizer: Tokenizer object (or function) for converting text into a tensor.

            normalize: Optional boolean flag for controlling if the outputs of the text encoder should be normalized.
                Default: `True`.

            batch_size: Optional batch size parameter used for computing text embeddings. Default: `64`.

            device: Optional string for specifying which device to use for computing embeddings. If omitted, it will be
                set to `'cuda'` if available else to `'cpu'`. Default: `None`.
        """
        self.txt_encoder = txt_encoder
        self.tokenizer = tokenizer
        self.normalize = normalize
        self.batch_size = batch_size

        if device:
            self.device = device
        else:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

    @torch.no_grad()
    def embedding_fn(self, text: list[str], disable_pbar: bool = False) -> torch.Tensor:
        """
        Generic method for computing text embeddings given a tokenizer and a text encoder model. Internally, called by
        the `IdealWords` class.

        Parameters
        ----------
            text: A list of string representations that will be tokenized and passed to the text encoder.

            disable_pbar: Optional boolean flag for disabling the tqdm progress bar.
        """
        self.txt_encoder.eval()
        self.txt_encoder.to(self.device)

        prompts = self.tokenizer(text)
        embeddings = []
        for batch in tqdm.tqdm(
            DataLoader(TensorDataset(prompts), batch_size=self.batch_size, shuffle=False),
            desc='Compute embeddings',
            disable=disable_pbar,
        ):
            embeddings.append(self.encode_text(batch[0].to(self.device)))
        embeddings = torch.cat(embeddings)

        return F.normalize(embeddings) if self.normalize else embeddings

    def encode_text(self, text: torch.Tensor) -> torch.Tensor:
        """
        Method that specifies how the text encoder should be used to compute embeddings. Overwrite this method if you
        don't want to use the forward method of your encoder to embed text.

        Parameters
        ----------
            text: Tensor containing the tokenized text representations of shape (batch_size, context_length)
        """
        return self.txt_encoder(text)

    def joint_repr(self, z: tuple[str, ...]) -> str:
        """
        Method that specifies how a tuple z in Z_1 x ... x Z_k should be represented as text. For example, the tuple
        `('blue', 'car')` might be represented as `'an image of a blue car'`. Overwrite this method if you don't want to
        simply join the tuple using spaces.

        Parameters
        ----------
            z: A tuple containing strings (one per factor) to be represented as a single string.
        """
        return ' '.join(z)

    def single_repr(self, zi: str, factor_idx: int) -> str:
        """
        Method that specifies how an element zi in Z_j should be represented as text. For example, the element
        `'blue'` might be represented as `'an image of a blue object'`. Overwrite this method if you don't want to
        simply use the element itself as a representation.

        Parameters
        ----------
            zi: A string representing an element of a specified factor Z_j.

            factor_idx: Zero-based index of the factor Z_j to which zi belongs.
        """
        _ = factor_idx
        return zi


class IdealWords:
    """
    Class for computing ideal word representations for a given factor embedding and factors. Optionally, weights for the
    individual elements of the factors can be specified.
    """

    def __init__(
        self,
        factor_embedding: FactorEmbedding,
        factors: list[list[str]],
        weights: list[list[float]] | None = None,
        score_mode: Literal['avg_dist', 'avg_sq_dist', 'paper_repro'] = 'avg_sq_dist',
        verbose: int = 1,
    ) -> None:
        """
        Parameters
        ----------
            factor_embedding: Factor embedding that handles text embedding and single/joint representation logic.

            factors: List of lists of strings where each list represents the elements of a factor.

            weights: Optional list of lists of floats where each list represents the weights for the elements of the
                factors. If omitted, uniform weights for all factor elements will be used. Default: `None`.

            score_mode: Optional flag to customize score computation.
                'avg_dist' - uses the average euclidean between embedding vectors for iw, rw, and avg score
                'avg_sq_dist' - uses the average squared euclidean between embedding vectors for iw, rw, and avg score
                'paper_repro' - uses average euclidean distance for iw and rw score and average squared euclidean
                    distance for avg score, empirically this seems to be what they used in their paper (Tables 6 and 7)
                In the paper, they state that they use squared distance everywhere which is why we choose this mode as
                the default. However, to reproduce the results from the paper, use the 'paper_repro' mode (for details
                see https://github.com/icetube23/ideal_words/issues/1). Default: `'avg_sq_dist'`.

            verbose: Optional int parameter to customize verbosity level.
                0 - show no tqdm progress bar at all
                1 - show tqdm progress bar during ideal word computation
                2 - show tqdm progress bar during ideal and real word computation
                Default: `1`.
        """
        # verify factors are disjoint
        assert len(reduce(lambda u, v: u | v, [set(factor) for factor in factors])) == sum(
            [len(factor) for factor in factors]
        ), 'Factors are expected to be disjoint.'

        self.factor_embedding = factor_embedding
        self.factors = factors
        self.pairs = list(product(*self.factors))
        self.pair2idx = {pair: i for i, pair in enumerate(self.pairs)}
        self.factor2idx = {zi: (i, j) for i, factor in enumerate(self.factors) for j, zi in enumerate(factor)}
        self.weights = weights if weights else [[1 / len(factor)] * len(factor) for factor in factors]
        self.device = self.factor_embedding.device

        assert score_mode in ['avg_dist', 'avg_sq_dist', 'paper_repro'], f'Invalid score mode: {score_mode}'
        self.score_mode = score_mode
        assert verbose in [0, 1, 2]
        self.verbose = verbose

        # verify weights
        assert len(self.factors) == len(self.weights)
        for factor, weight in zip(self.factors, self.weights):
            assert len(factor) == len(weight)
            assert isclose(sum(weight), 1)

        # precompute indices over which we need to average later
        self.factor_indices = [{zi: [] for zi in factor} for factor in self.factors]
        for i, z in enumerate(self.pairs):
            for j, zi in enumerate(z):
                self.factor_indices[j][zi].append(i)

        # compute embeddings and ideal words
        self._compute_ideal_words()

        # approximation cache
        self._uz: dict[str, torch.Tensor] = {}

        # lazily computed real words
        self.real_words = None

        # lazily computed compositionality scores from the paper
        self._iw_score = None
        self._rw_score = None
        self._avg_score = None

        # lazily computed custom compositionality scores
        self._iw_accuracy = None
        self._rw_accuracy = None

    @torch.no_grad()
    def _compute_ideal_words(self) -> None:
        """Compute ideal words as in Equation 4 of the paper."""
        # expand factor / weight lists into paired representation to allow for tensor operations
        alphas = torch.tensor(list(product(*self.weights))).to(self.device).double()
        betas = alphas.prod(dim=1).unsqueeze(-1).double()

        # compute embeddings for each combination of factors
        captions = [self.factor_embedding.joint_repr(pair) for pair in self.pairs]
        embeddings = self.factor_embedding.embedding_fn(captions, disable_pbar=self.verbose < 1).double()

        # u_zero is a weighted average of all embeddings
        u_zero = (embeddings * betas).sum(dim=0)

        ideal_words = []
        for i, factor in enumerate(self.factors):
            # precompute weighted embeddings instead
            weighted_embeddings = alphas[:, i].reciprocal().unsqueeze(-1) * betas * embeddings

            # compute u_zi for each zi in factor Zi
            u_zi = []
            for zi in factor:
                inds = self.factor_indices[i][zi]
                u_zi.append(weighted_embeddings[inds].sum(dim=0))
            u_zi = torch.stack(u_zi) - u_zero
            ideal_words.append(u_zi.float())

        # we use double precision for the weighted averages but afterwards it is not really needed anymore
        self.embeddings = embeddings.float()
        self.u_zero = u_zero.float()
        self.ideal_words = ideal_words

    @torch.no_grad()
    def _compute_real_words(self) -> None:
        """Compute real words by encoding factor elements individually."""
        # real words are computed by embedding a prompt containing only info from a single factor at a time
        real_words = []
        for idx, factor in enumerate(self.factors):
            captions = [self.factor_embedding.single_repr(zi, idx) for zi in factor]
            embeddings = self.factor_embedding.embedding_fn(captions, disable_pbar=self.verbose < 2).float()
            real_words.append(embeddings)

        self.real_words = real_words

    @torch.no_grad()
    def _materialize_uz(self, approx: str) -> torch.Tensor:
        # if not already materialized and cached, compute approximations
        if approx not in self._uz:
            uz_hat = torch.stack([self.get_uz(pair, approx=approx) for pair in self.pairs])
            assert uz_hat.shape == self.embeddings.shape
            self._uz[approx] = uz_hat

        return self._uz[approx]

    def get_iw(self, zi: str) -> torch.Tensor:
        """Retrieve the ideal word representation of a given factor element."""
        assert zi in self.factor2idx, f'Unknown concept: {zi}'
        i, j = self.factor2idx[zi]
        return self.ideal_words[i][j]

    def get_rw(self, zi: str) -> torch.Tensor:
        """Retrieve the real word representation of a given factor element."""
        if self.real_words is None:
            self._compute_real_words()
            assert self.real_words is not None

        assert zi in self.factor2idx, f'Unknown concept: {zi}'
        i, j = self.factor2idx[zi]
        return self.real_words[i][j]

    def get_uz(self, z: Sequence[str], approx: str = 'ideal') -> torch.Tensor:
        """Compute an approximation to the actual embedding of tuple z using either ideal or real words."""
        assert len(z) == len(self.factors)

        # approximations are already cached and need not to be recomputed
        if approx in self._uz:
            idx = self.pair2idx[tuple(z)]
            return self._uz[approx][idx]

        if approx == 'ideal':
            return torch.stack([self.get_iw(zi) for zi in z]).sum(dim=0) + self.u_zero
        elif approx == 'real':
            return torch.stack([self.get_rw(zi) for zi in z]).mean(dim=0)
        else:
            raise ValueError(f'Invalid approximation mode: {approx}')

    def _score(self, approx: str) -> tuple[float, float]:
        # get word approximation for each combination of factors
        uz_hat = self._materialize_uz(approx)

        # compute distances between the compositional approximations and the actual embeddings
        dists = torch.linalg.vector_norm(self.embeddings - uz_hat, dim=1)

        # use average squared distance for score
        if self.score_mode == 'avg_sq_dist':
            dists.square_()

        return dists.mean().cpu().item(), dists.std().cpu().item()

    @property
    def iw_score(self) -> tuple[float, float]:
        """
        Retrieve the ideal words score, i.e., the average distance between the actual embeddings and the ideal word
        approximations. The first time this property is accessed the score will be computed and will be cached for
        successive accesses.
        """
        if self._iw_score is None:
            self._iw_score = self._score('ideal')

        return self._iw_score

    @property
    def rw_score(self) -> tuple[float, float]:
        """
        Retrieve the real words score, i.e., the average distance between the actual embeddings and the real word
        approximations. The first time this property is accessed the score will be computed and will be cached for
        successive accesses.
        """
        if self._rw_score is None:
            self._rw_score = self._score('real')

        return self._rw_score

    @property
    def avg_score(self) -> tuple[float, float]:
        """
        Compute the average pairwise distances between the actual embeddings. The first time this property is accessed
        the score will be computed and will be cached for successive accesses.
        """
        if self._avg_score is None:
            # compute pairwise distances of original embedding vectors
            embeddings = self.embeddings.half()  # save some memory as this metric is expensive to compute
            dists = torch.cdist(embeddings, embeddings, compute_mode='use_mm_for_euclid_dist')

            # dists is a symmetric matrix with diagonal 0 because cdist considers ordered pairs
            # so we only average over the upper triangular half of dists
            dists = dists[torch.ones_like(dists, dtype=torch.bool).triu(1)]

            # use average squared distance for score
            if self.score_mode == 'avg_sq_dist' or self.score_mode == 'paper_repro':
                dists.square_()

            self._avg_score = dists.mean().cpu().item(), dists.std().cpu().item()

        return self._avg_score

    def _accuracy(self, approx: str) -> float:
        # get word approximation for each combination of factors
        uz_hat = self._materialize_uz(approx).half()
        embeddings = self.embeddings.half()  # save some memory as this metric is expensive to comput

        # compute pairwise distances between the approximations and the actual embeddings
        dists = torch.cdist(embeddings, uz_hat, compute_mode='use_mm_for_euclid_dist')

        # find nearest neighbor for each approximated embedding and compare against expected nearest neighbors
        approx_matches = dists.argmin(dim=0)
        expected_matches = torch.arange(len(approx_matches), device=self.device)

        return (approx_matches == expected_matches).float().mean().cpu().item()

    @property
    def iw_accuracy(self) -> float:
        """
        Retrieve the ideal words accuracy, i.e., find the nearest actual embedding vector for each approximate embedding
        vector and compute the accuracy over how often the nearest neighbor matches the expected actual emebdding
        vector. The first time this property is accessed the score will be computed and will be cached for successive
        accesses.
        """
        if self._iw_accuracy is None:
            self._iw_accuracy = self._accuracy('ideal')

        return self._iw_accuracy

    @property
    def rw_accuracy(self) -> float:
        """
        Retrieve the real words accuracy, i.e., find the nearest actual embedding vector for each approximate embedding
        vector and compute the accuracy over how often the nearest neighbor matches the expected actual emebdding
        vector. The first time this property is accessed the score will be computed and will be cached for successive
        accesses.
        """
        if self._rw_accuracy is None:
            self._rw_accuracy = self._accuracy('real')

        return self._rw_accuracy
